"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.lambdaRequestTracker = void 0;
const context_1 = require("./context");
const AMAZON_TRACE_ID = '_X_AMZN_TRACE_ID';
const CORRELATION_HEADER = 'x-correlation-';
const CORRELATION_ID = `${CORRELATION_HEADER}id`;
const CORRELATION_TRACE_ID = `${CORRELATION_HEADER}trace-id`;
/**
 * Creates a function for tracing Lambda request context across logging calls
 * @param options The request options
 */
const lambdaRequestTracker = (options = {}) => (event, context) => {
    var _a;
    const ctx = {
        awsRequestId: context.awsRequestId,
    };
    // capture api gateway request ID
    const apiRequestId = (_a = event.requestContext) === null || _a === void 0 ? void 0 : _a.requestId;
    if (apiRequestId) {
        ctx.apiRequestId = apiRequestId;
    }
    // capture any correlation headers sent from upstream callers
    if (event.headers) {
        Object.keys(event.headers).forEach((header) => {
            if (header.toLowerCase().startsWith(CORRELATION_HEADER)) {
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                ctx[header] = event.headers[header];
            }
        });
    }
    // capture the xray trace id if its enabled
    if (process.env[AMAZON_TRACE_ID]) {
        ctx[CORRELATION_TRACE_ID] = process.env[AMAZON_TRACE_ID];
    }
    // set the correlation id if not already set by upstream callers
    if (!ctx[CORRELATION_ID]) {
        ctx[CORRELATION_ID] = context.awsRequestId;
    }
    // handle custom request level mixins
    if (options.requestMixin) {
        const result = options.requestMixin(event, context);
        for (const key in result) {
            // Cast this to string for typescript
            // when the JSON serializer runs, by default it omits undefined properties
            ctx[key] = result[key];
        }
    }
    const storageProvider = options.storageProvider || context_1.GlobalContextStorageProvider;
    if (storageProvider) {
        storageProvider.setContext(ctx);
    }
};
exports.lambdaRequestTracker = lambdaRequestTracker;
